@@ -18,6 +18,7 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
+from __future__ import print_function
 import sip
 
 sip.setapi("QString", 2)
@@ -25,7 +26,7 @@ sip.setapi("QVariant", 2)
 
 from PyQt4 import QtGui, QtCore
 
-import getopt
+import argparse
 import os.path
 import platform
 import re
@@ -48,11 +49,9 @@ _orig_shutil_copystat = shutil.copystat
 shutil.copystat = _patched_shutil_copystat
 
 import picard.resources
-import picard.plugins
 from picard.i18n import setup_gettext
 
-from picard import (PICARD_APP_NAME, PICARD_ORG_NAME,
-                    PICARD_FANCY_VERSION_STR, __version__,
+from picard import (PICARD_APP_NAME, PICARD_ORG_NAME, PICARD_FANCY_VERSION_STR,
                     log, acoustid, config)
 from picard.album import Album, NatAlbum
 from picard.browser.browser import BrowserIntegration
@@ -67,6 +66,7 @@ from picard.track import Track, NonAlbumTrack
 from picard.releasegroup import ReleaseGroup
 from picard.collection import load_user_collections
 from picard.ui.mainwindow import MainWindow
+from picard.ui.itemviews import BaseTreeView
 from picard.plugin import PluginManager
 from picard.acoustidmanager import AcoustIDManager
 from picard.config_upgrade import upgrade_config
@@ -81,6 +81,11 @@ from picard.util import (
     versions,
 )
 from picard.webservice import XmlWebService
+from picard.ui.searchdialog import (
+    TrackSearchDialog,
+    AlbumSearchDialog,
+    ArtistSearchDialog
+)
 
 
 class Tagger(QtGui.QApplication):
@@ -94,11 +99,15 @@ class Tagger(QtGui.QApplication):
 
     __instance = None
 
-    def __init__(self, args, localedir, autoupdate, debug=False):
-        QtGui.QApplication.__init__(self, args)
+    def __init__(self, picard_args, unparsed_args, localedir, autoupdate):
+        # Set the WM_CLASS to 'MusicBrainz-Picard' so desktop environments
+        # can use it to look up the app
+        QtGui.QApplication.__init__(self, ['MusicBrainz-Picard'] + unparsed_args)
         self.__class__.__instance = self
 
-        self._args = args
+        config._setup(self, picard_args.config_file)
+
+        self._cmdline_files = picard_args.FILE
         self._autoupdate = autoupdate
         self._debug = False
 
@@ -135,15 +144,12 @@ class Tagger(QtGui.QApplication):
             signal.signal(signal.SIGTERM, self.signal)
 
         # Setup logging
-        self.debug(debug or "PICARD_DEBUG" in os.environ)
+        self.debug(picard_args.debug or "PICARD_DEBUG" in os.environ)
         log.debug("Starting Picard from %r", os.path.abspath(__file__))
         log.debug("Platform: %s %s %s", platform.platform(),
                   platform.python_implementation(), platform.python_version())
         log.debug("Versions: %s", versions.as_string())
-        if config.storage_type == config.REGISTRY_PATH:
-            log.debug("Configuration registry path: %s", config.storage)
-        else:
-            log.debug("Configuration file path: %s", config.storage)
+        log.debug("Configuration file path: %r", config.config.fileName())
 
         # TODO remove this before the final release
         if sys.platform == "win32":
@@ -152,12 +158,12 @@ class Tagger(QtGui.QApplication):
             olduserdir = "~/.picard"
         olduserdir = os.path.expanduser(olduserdir)
         if os.path.isdir(olduserdir):
-            log.info("Moving %s to %s", olduserdir, USER_DIR)
+            log.info("Moving %r to %r", olduserdir, USER_DIR)
             try:
                 shutil.move(olduserdir, USER_DIR)
             except:
                 pass
-        log.debug("User directory: %s", os.path.abspath(USER_DIR))
+        log.debug("User directory: %r", os.path.abspath(USER_DIR))
 
         # for compatibility with pre-1.3 plugins
         QtCore.QObject.tagger = self
@@ -192,6 +198,7 @@ class Tagger(QtGui.QApplication):
         if not os.path.exists(USER_PLUGIN_DIR):
             os.makedirs(USER_PLUGIN_DIR)
         self.pluginmanager.load_plugindir(USER_PLUGIN_DIR)
+        self.pluginmanager.query_available_plugins()
 
         self.acoustidmanager = AcoustIDManager()
         self.browser_integration = BrowserIntegration()
@@ -205,6 +212,7 @@ class Tagger(QtGui.QApplication):
         self.nats = None
         self.window = MainWindow()
         self.exit_cleanup = []
+        self.stopping = False
 
     def register_cleanup(self, func):
         self.exit_cleanup.append(func)
@@ -260,26 +268,27 @@ class Tagger(QtGui.QApplication):
             self.nats.update()
 
     def exit(self):
-        log.debug("exit")
+        log.debug("Picard stopping")
         self.stopping = True
         self._acoustid.done()
         self.thread_pool.waitForDone()
+        self.save_thread_pool.waitForDone()
         self.browser_integration.stop()
         self.xmlws.stop()
-        for f in self.exit_cleanup:
-            f()
+        self.run_cleanup()
+        QtCore.QCoreApplication.processEvents()
 
     def _run_init(self):
-        if self._args:
+        if self._cmdline_files:
             files = []
-            for file in self._args:
+            for file in self._cmdline_files:
                 if os.path.isdir(file):
                     self.add_directory(decode_filename(file))
                 else:
                     files.append(decode_filename(file))
             if files:
                 self.add_files(files)
-            del self._args
+            del self._cmdline_files
 
     def run(self):
         if config.setting["browser_integration"]:
@@ -324,6 +333,9 @@ class Tagger(QtGui.QApplication):
                 self.analyze([file])
 
     def move_files(self, files, target):
+        if target is None:
+            log.debug("Aborting move since target is invalid")
+            return
         if isinstance(target, (Track, Cluster)):
             for file in files:
                 file.move(target)
@@ -346,10 +358,10 @@ class Tagger(QtGui.QApplication):
         for filename in filenames:
             filename = os.path.normpath(os.path.realpath(filename))
             if ignore_hidden and is_hidden(filename):
-                log.debug("File ignored (hidden): %s" % (filename))
+                log.debug("File ignored (hidden): %r" % (filename))
                 continue
             if ignoreregex is not None and ignoreregex.search(filename):
-                log.info("File ignored (matching %s): %s" % (pattern, filename))
+                log.info("File ignored (matching %r): %r" % (pattern, filename))
                 continue
             if filename not in self.files:
                 file = open_file(filename)
@@ -366,12 +378,18 @@ class Tagger(QtGui.QApplication):
                 file.load(partial(self._file_loaded, target=target))
 
     def add_directory(self, path):
+        if config.setting['recursively_add_files']:
+            self._add_directory_recursive(path)
+        else:
+            self._add_directory_non_recursive(path)
+
+    def _add_directory_recursive(self, path):
         ignore_hidden = config.setting["ignore_hidden_files"]
         walk = os.walk(unicode(path))
 
         def get_files():
             try:
-                root, dirs, files = walk.next()
+                root, dirs, files = next(walk)
                 if ignore_hidden:
                     dirs[:] = [d for d in dirs if not is_hidden(os.path.join(root, d))]
             except StopIteration:
@@ -383,7 +401,7 @@ class Tagger(QtGui.QApplication):
                         'count': number_of_files,
                         'directory': root,
                     }
-                    log.debug("Adding %(count)d files from '%(directory)s'" %
+                    log.debug("Adding %(count)d files from '%(directory)r'" %
                               mparms)
                     self.window.set_statusbar_message(
                         ungettext(
@@ -404,21 +422,71 @@ class Tagger(QtGui.QApplication):
 
         process(True, False)
 
+    def _add_directory_non_recursive(self, path):
+        files = []
+        for f in os.listdir(path):
+            listing = os.path.join(path, f)
+            if os.path.isfile(listing):
+                files.append(listing)
+        number_of_files = len(files)
+        if number_of_files:
+            mparms = {
+                'count': number_of_files,
+                'directory': path,
+            }
+            log.debug("Adding %(count)d files from '%(directory)r'" %
+                      mparms)
+            self.window.set_statusbar_message(
+                ungettext(
+                    "Adding %(count)d file from '%(directory)s' ...",
+                    "Adding %(count)d files from '%(directory)s' ...",
+                    number_of_files),
+                mparms,
+                translate=None,
+                echo=None
+            )
+            # Function call only if files exist
+            self.add_files(files)
+
     def get_file_lookup(self):
         """Return a FileLookup object."""
         return FileLookup(self, config.setting["server_host"],
                           config.setting["server_port"],
                           self.browser_integration.port)
 
+    def copy_files(self, objects):
+        mimeData = QtCore.QMimeData()
+        mimeData.setUrls([QtCore.QUrl.fromLocalFile(f.filename) for f in (self.get_files_from_objects(objects))])
+        self.clipboard().setMimeData(mimeData)
+
+    def paste_files(self, target):
+        mimeData = self.clipboard().mimeData()
+        if mimeData.hasUrls():
+            BaseTreeView.drop_urls(mimeData.urls(), target)
+
     def search(self, text, type, adv=False):
         """Search on the MusicBrainz website."""
         lookup = self.get_file_lookup()
-        getattr(lookup, type + "Search")(text, adv)
+        if config.setting["builtin_search"]:
+            if type == "track" and not lookup.mbidLookup(text, 'recording'):
+                dialog = TrackSearchDialog(self.window)
+                dialog.search(text)
+                dialog.exec_()
+            elif type == "album" and not lookup.mbidLookup(text, 'release'):
+                dialog = AlbumSearchDialog(self.window)
+                dialog.search(text)
+                dialog.exec_()
+            elif type == "artist" and not lookup.mbidLookup(text, 'artist'):
+                dialog = ArtistSearchDialog(self.window)
+                dialog.search(text)
+                dialog.exec_()
+        else:
+            getattr(lookup, type + "Search")(text, adv)
 
     def collection_lookup(self):
         """Lookup the users collections on the MusicBrainz website."""
         lookup = self.get_file_lookup()
-        lookup.collectionLookup(config.setting["username"])
+        lookup.collectionLookup(config.persist["oauth_username"])
 
     def browser_lookup(self, item):
         """Lookup the object's metadata on the MusicBrainz website."""
@@ -536,6 +604,8 @@ class Tagger(QtGui.QApplication):
                     }
                 )
                 self.remove_album(obj)
+            elif isinstance(obj, UnmatchedFiles):
+                files.extend(list(obj.files))
             elif isinstance(obj, Cluster):
                 self.remove_cluster(obj)
         if files:
@@ -635,6 +705,11 @@ class Tagger(QtGui.QApplication):
             if obj.can_refresh():
                 obj.load(priority=True, refresh=True)
 
+    def bring_tagger_front(self):
+        self.window.setWindowState(self.window.windowState() & ~QtCore.Qt.WindowMinimized | QtCore.Qt.WindowActive)
+        self.window.raise_()
+        self.window.activateWindow()
+
     @classmethod
     def instance(cls):
         return cls.__instance
@@ -652,21 +727,30 @@ class Tagger(QtGui.QApplication):
         self.signalnotifier.setEnabled(True)
 
 
-def help():
-    print """Usage: %s [OPTIONS] [FILE] [FILE] ...
-
-If one of the filenames begins with a hyphen, use -- to separate the options
-from the filenames.
-
-Options:
-    -d, --debug             enable debug-level logging
-    -h, --help              display this help and exit
-    -v, --version           display version information and exit
-""" % (sys.argv[0],)
-
-
 def version():
-    print "%s %s %s" % (PICARD_ORG_NAME, PICARD_APP_NAME, PICARD_FANCY_VERSION_STR)
+    print("%s %s %s" % (PICARD_ORG_NAME, PICARD_APP_NAME, PICARD_FANCY_VERSION_STR))
+
+
+def longversion():
+    print(versions.as_string())
+
+
+def process_picard_args():
+    parser = argparse.ArgumentParser(
+        epilog="If one of the filenames begins with a hyphen, use -- to separate the options from the filenames."
+    )
+    parser.add_argument("-c", "--config-file", action='store',
+                        default=None,
+                        help="location of the configuration file")
+    parser.add_argument("-d", "--debug", action='store_true',
+                        help="enable debug-level logging")
+    parser.add_argument('-v', '--version', action='store_true',
+                        help="display version information and exit")
+    parser.add_argument("-V", "--long-version", action='store_true',
+                        help="display long version information and exit")
+    parser.add_argument('FILE', nargs='*')
+    picard_args, unparsed_args = parser.parse_known_args()
+    return picard_args, unparsed_args
 
 
 def main(localedir=None, autoupdate=True):
@@ -675,15 +759,13 @@ def main(localedir=None, autoupdate=True):
     QtGui.QApplication.setOrganizationName(PICARD_ORG_NAME)
 
     signal.signal(signal.SIGINT, signal.SIG_DFL)
-    opts, args = getopt.gnu_getopt(sys.argv[1:], "hvd", ["help", "version", "debug"])
-    kwargs = {}
-    for opt, arg in opts:
-        if opt in ("-h", "--help"):
-            return help()
-        elif opt in ("-v", "--version"):
-            return version()
-        elif opt in ("-d", "--debug"):
-            kwargs["debug"] = True
-    tagger = Tagger(args, localedir, autoupdate, **kwargs)
+
+    picard_args, unparsed_args = process_picard_args()
+    if picard_args.version:
+        return version()
+    if picard_args.long_version:
+        return longversion()
+
+    tagger = Tagger(picard_args, unparsed_args, localedir, autoupdate)
     tagger.startTimer(1000)
     sys.exit(tagger.run())
